#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# type: ignore

"""Thoth-adviser CLI."""

import json
import logging
import os
import pickle
import random
import sys
import time
from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

import attr
import click
import yaml
import termial_random
from thoth.analyzer import print_command_result
from thoth.common import init_logging
from thoth.common import RuntimeEnvironment
from thoth.python import Constraints
from thoth.python import Pipfile
from thoth.python import PipfileLock
from thoth.python import Project
from thoth.python.exceptions import UnsupportedConfigurationError
from prometheus_client import CollectorRegistry
from prometheus_client import Gauge
from prometheus_client import push_to_gateway

from thoth.adviser.dependency_monkey import DependencyMonkey
from thoth.adviser.digests_fetcher import GraphDigestsFetcher
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import PythonRecommendationOutput
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import AdviserException
from thoth.adviser.exceptions import InternalError
from thoth.adviser.pipeline_builder import PipelineBuilder
from thoth.adviser.prescription import Prescription
from thoth.adviser import Resolver
from thoth.adviser import __title__ as analyzer_name
from thoth.adviser import __version__ as analyzer_version
from thoth.adviser.run import subprocess_run
import thoth.adviser.predictors as predictors

init_logging()

_LOGGER = logging.getLogger("thoth.adviser")


prometheus_registry = CollectorRegistry()

_THOTH_DEPLOYMENT_NAME = os.getenv("THOTH_DEPLOYMENT_NAME")
_THOTH_METRICS_PUSHGATEWAY_URL = os.getenv("PROMETHEUS_PUSHGATEWAY_URL")
_DEFAULT_PLATFORM = "linux-x86_64"


_METRIC_INFO = Gauge(
    "thoth_adviser_info",
    "Thoth adviser information",
    ["env", "version"],
    registry=prometheus_registry,
)


_METRIC_DATABASE_SCHEMA_SCRIPT = Gauge(
    "thoth_database_schema_revision_script",
    "Thoth database schema revision from script",
    ["component", "revision", "env"],
    registry=prometheus_registry,
)


@attr.s(slots=True)
class _PrintFunc:
    """A print function - a workaround for typing and kwargs arguments."""

    func = attr.ib(type=Callable)

    def __call__(self, duration: float, result: Dict[str, Any]) -> None:
        self.func(duration=duration, result=result)


def _print_version(ctx: click.Context, _, value: str):
    """Print adviser version and exit."""
    if not value or ctx.resilient_parsing:
        return

    click.echo(analyzer_version)
    ctx.exit()


def _instantiate_project(
    requirements: str,
    requirements_locked: Optional[str] = None,
    *,
    runtime_environment: RuntimeEnvironment = None,
    constraints: Optional[str] = None,
):
    """Create Project instance based on arguments passed to CLI."""
    try:
        with open(requirements, "r") as requirements_file:
            requirements = requirements_file.read()
    except (OSError, FileNotFoundError):
        # We we gather values from env vars, un-escape new lines.
        requirements = requirements.replace("\\n", "\n")

    if requirements_locked:
        try:
            with open(requirements_locked, "r") as requirements_file:
                requirements_locked = requirements_file.read()
        except (OSError, FileNotFoundError):
            # We we gather values from env vars, un-escape new lines.
            requirements_locked = requirements_locked.replace("\\n", "\n")

    pipfile = Pipfile.from_string(requirements)
    pipfile_lock = None
    if requirements_locked and requirements_locked != "null":
        pipfile_lock = PipfileLock.from_string(requirements_locked, pipfile)

    constraints_instance = None
    if constraints:
        try:
            with open(constraints, "r") as constraints_file:
                constraints_content = constraints_file.read()
        except (OSError, FileNotFoundError):
            # We we gather values from env vars, un-escape new lines.
            constraints_content = constraints.replace("\\n", "\n")

        try:
            constraints_instance = Constraints.from_dict(json.loads(constraints_content))
        except json.decoder.JSONDecodeError:
            constraints_instance = Constraints.from_string(constraints_content)

    runtime_environment = runtime_environment or RuntimeEnvironment.from_dict({})
    if not runtime_environment.platform:
        runtime_environment.platform = _DEFAULT_PLATFORM

    project = Project(
        pipfile=pipfile,
        pipfile_lock=pipfile_lock,
        runtime_environment=runtime_environment,
        constraints=constraints_instance or Constraints(),
    )

    return project


def _get_adviser_predictor(predictor: str, recommendation_type: RecommendationType) -> Tuple[type, Dict[str, Any]]:
    """Get adviser predictor based on command line option."""
    if predictor != "AUTO":
        return getattr(predictors, predictor), {}

    if recommendation_type == RecommendationType.LATEST:
        return predictors.ApproximatingLatest, {}
    elif (
        recommendation_type == RecommendationType.STABLE
        or recommendation_type == RecommendationType.TESTING
        or recommendation_type == RecommendationType.PERFORMANCE
        or recommendation_type == RecommendationType.SECURITY
    ):
        return (
            predictors.TemporalDifference,
            {
                "step": 1,
                "temperature_coefficient": predictors.TemporalDifference.obtain_default_configuration(
                    "temperature_coefficient"
                ),
                "trace": False,
            },
        )

    raise ValueError(f"Unknown recommendation type: {recommendation_type!r}")


def _get_dependency_monkey_predictor(predictor: str, decision_type: DecisionType) -> type:
    """Get dependency monkey predictor based on command line option."""
    if predictor != "AUTO":
        return getattr(predictors, predictor)

    if decision_type == DecisionType.RANDOM:
        return predictors.RandomWalk
    elif decision_type == DecisionType.ALL:
        return predictors.ApproximatingLatest

    raise ValueError(f"Unknown decision type: {decision_type!r}")


def _get_predictor_kwargs(predictor_config: Optional[str]) -> Dict[str, Any]:
    """Get kwargs for a predictor instance."""
    if predictor_config is None:
        return {}

    if os.path.isfile(predictor_config):
        with open(predictor_config, "r") as predictor_config_file:
            return yaml.safe_load(predictor_config_file)

    return yaml.safe_load(predictor_config)


@click.group()
@click.pass_context
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    envvar="THOTH_ADVISER_DEBUG",
    help="Be verbose about what's going on.",
)
@click.option(
    "--version",
    is_flag=True,
    is_eager=True,
    callback=_print_version,
    expose_value=False,
    help="Print adviser version and exit.",
)
@click.option(
    "--metadata",
    type=str,
    envvar="THOTH_ADVISER_METADATA",
    help="Metadata in a form of a JSON which are used for carrying additional context in Thoth deployment.",
)
def cli(ctx=None, verbose=False, metadata=None):
    """Thoth adviser command line interface."""
    if ctx:
        ctx.auto_envvar_prefix = "THOTH_ADVISER"

    if verbose:
        _LOGGER.setLevel(logging.DEBUG)

    _LOGGER.debug("Debug mode is on")
    _LOGGER.info("Version: %s", analyzer_version)

    # This value is unused here, but is reported from click context.
    metadata = metadata


@cli.command()
@click.pass_context
@click.option(
    "--requirements",
    "-r",
    type=str,
    envvar="THOTH_ADVISER_REQUIREMENTS",
    required=True,
    help="Pipfile to be checked for provenance.",
)
@click.option(
    "--requirements-locked",
    "-l",
    type=str,
    envvar="THOTH_ADVISER_REQUIREMENTS_LOCKED",
    required=True,
    help="Pipenv.lock file stating currently locked packages.",
)
@click.option(
    "--output",
    "-o",
    type=str,
    envvar="THOTH_ADVISER_OUTPUT",
    default="-",
    help="Output file or remote API to print results to, in case of URL a POST request is issued.",
)
@click.option("--no-pretty", "-P", is_flag=True, help="Do not print results nicely.")
@click.option(
    "--whitelisted-sources",
    "-i",
    type=str,
    required=False,
    envvar="THOTH_WHITELISTED_SOURCES",
    help="A comma separated list of whitelisted simple repositories providing packages - if not "
    "provided, all indexes are whitelisted (example: https://pypi.org/simple/).",
)
def provenance(
    click_ctx: click.Context,
    requirements: str,
    requirements_locked: str,
    output: str,
    whitelisted_sources: Optional[str] = None,
    no_pretty: bool = False,
):
    """Check provenance of packages based on configuration."""
    parameters = locals()
    parameters.pop("click_ctx")
    start_time = time.monotonic()
    _LOGGER.debug("Passed arguments: %s", parameters)

    whitelisted_sources = whitelisted_sources.split(",") if whitelisted_sources else []
    result = {
        "error": None,
        "report": [],
        "parameters": {"whitelisted_indexes": whitelisted_sources, "project": None},
        "input": None,
    }
    try:
        project = _instantiate_project(requirements, requirements_locked)
        result["parameters"]["project"] = project.to_dict()
        report = project.check_provenance(
            whitelisted_sources=whitelisted_sources,
            digests_fetcher=GraphDigestsFetcher(),
        )
    except (AdviserException, UnsupportedConfigurationError) as exc:
        if isinstance(exc, InternalError):
            # Re-raise internal exceptions that shouldn't occur here.
            raise

        _LOGGER.exception("Error during checking provenance: %s", str(exc))
        result["error"] = True
        result["error_msg"] = str(exc)
        result["report"] = [{"type": "ERROR", "justification": f"{str(exc)} ({type(exc).__name__})"}]
    else:
        result["error"] = False
        result["error_msg"] = None
        result["report"] = report

    print_command_result(
        click_ctx,
        result,
        analyzer=analyzer_name,
        analyzer_version=analyzer_version,
        output=output,
        duration=time.monotonic() - start_time,
        pretty=not no_pretty,
    )

    click_ctx.exit(int(result["error"] is True))


@cli.command()
@click.pass_context
@click.option(
    "--requirements",
    "-r",
    type=str,
    envvar="THOTH_ADVISER_REQUIREMENTS",
    required=True,
    help="Requirements to be advised.",
)
@click.option(
    "--requirements-locked",
    "-l",
    type=str,
    envvar="THOTH_ADVISER_REQUIREMENTS_LOCKED",
    help="Currently locked down requirements used.",
)
@click.option(
    "--requirements-format",
    "-f",
    envvar="THOTH_REQUIREMENTS_FORMAT",
    default="pipenv",
    required=True,
    type=click.Choice(["pipenv", "requirements"]),
    help="The output format of requirements that are computed based on recommendations.",
)
@click.option(
    "--output",
    "-o",
    type=str,
    envvar="THOTH_ADVISER_OUTPUT",
    default="-",
    help="Output file or remote API to print results to, in case of URL a POST request is issued.",
)
@click.option("--no-pretty", "-P", is_flag=True, help="Do not print results nicely.")
@click.option(
    "--recommendation-type",
    "-t",
    envvar="THOTH_ADVISER_RECOMMENDATION_TYPE",
    default="stable",
    required=True,
    type=click.Choice([e.name.lower() for e in RecommendationType]),
    help="Type of recommendation generated based on knowledge base.",
)
@click.option(
    "--count",
    type=int,
    envvar="THOTH_ADVISER_COUNT",
    help="Number of software stacks shown in the output.",
    default=Resolver.DEFAULT_COUNT,
    show_default=True,
)
@click.option(
    "--limit",
    type=int,
    envvar="THOTH_ADVISER_LIMIT",
    help="Number of software stacks that should be taken into account (stop after reaching the limit).",
    default=Resolver.DEFAULT_LIMIT,
    show_default=True,
)
@click.option(
    "--seed",
    type=int,
    envvar="THOTH_ADVISER_SEED",
    help="A seed used for generating random numbers (defaults to time if omitted).",
)
@click.option(
    "--library-usage",
    type=str,
    envvar="THOTH_ADVISER_LIBRARY_USAGE",
    help="Add library usage information to adviser's resolution algorithm.",
)
@click.option(
    "--runtime-environment",
    "-e",
    envvar="THOTH_ADVISER_RUNTIME_ENVIRONMENT",
    type=str,
    help="Runtime environment specification (file or directly JSON) to describe target environment.",
)
@click.option("--plot", envvar="THOTH_ADVISER_PLOT", type=str, help="Plot history of predictor.")
@click.option(
    "--beam-width",
    "-b",
    envvar="THOTH_ADVISER_BEAM_WIDTH",
    type=int,
    default=Resolver.DEFAULT_BEAM_WIDTH,
    help="Width of the beam used.",
)
@click.option(
    "--limit-latest-versions",
    envvar="THOTH_ADVISER_LIMIT_LATEST_VERSIONS",
    type=int,
    default=Resolver.DEFAULT_LIMIT_LATEST_VERSIONS,
    help="Limit number of latest versions considered for dependency graphs.",
)
@click.option(
    "--predictor",
    envvar="THOTH_ADVISER_PREDICTOR",
    default="AUTO",
    type=click.Choice(predictors.__all__ + ["AUTO"]),
    help="Predictor to be used with the resolver, select the most appropriate one in case of AUTO.",
)
@click.option(
    "--predictor-config",
    envvar="THOTH_ADVISER_PREDICTOR_CONFIG",
    default=None,
    type=str,
    metavar="CONFIG",
    help="Predictor configuration - passed as a path to YAML file or as a YAML string.",
)
@click.option(
    "--pipeline",
    envvar="THOTH_ADVISER_PIPELINE",
    default=None,
    type=str,
    metavar="PIPELINE",
    help="Pipeline configuration supplied in a form of JSON/YAML or a path to a file, "
    "disjoint with pipeline prescription.",
)
@click.option(
    "--prescription",
    envvar="THOTH_ADVISER_PRESCRIPTION",
    default=None,
    type=str,
    multiple=True,
    metavar="PRESCRIPTION",
    help="Pipeline prescription supplied in a form of JSON/YAML or a path to a file, "
    "disjoint with pipeline configuration. Multiple files can be supplied by using `,' as a delimiter.",
)
@click.option(
    "--constraints",
    envvar="THOTH_ADVISER_CONSTRAINTS",
    default=None,
    type=str,
    metavar="CONSTRAINTS.txt",
    help="Constraints to be used during the resolution.",
)
@click.option(
    "--labels",
    "-l",
    envvar="THOTH_ADVISER_LABELS",
    type=str,
    metavar='{"key1": "val1", "key2": "val2"}',
    default=None,
    show_default=True,
    help="Labels used used to label the request.",
)
@click.option(
    "--user-stack-scoring/--no-user-stack-scoring",
    envvar="THOTH_ADVISER_USER_STACK_SCORING",
    is_flag=True,
    default=True,
    show_default=True,
    help="Turn off or on user stack scoring - the lock file supplied, if any, will be used "
    "as a base for relative stack quality comparision. Adviser will score the supplied "
    "lock file and will try to find a better stack.",
)
@click.option(
    "--dev/--no-dev",
    envvar="THOTH_ADVISER_DEV",
    is_flag=True,
    default=False,
    show_default=True,
    help="Consider or do not consider development dependencies during resolution.",
)
def advise(
    click_ctx: click.Context,
    *,
    beam_width: int,
    count: int,
    limit: int,
    output: str,
    recommendation_type: str,
    requirements_format: str,
    requirements: str,
    predictor: str,
    predictor_config: Optional[str] = None,
    library_usage: Optional[str] = None,
    limit_latest_versions: Optional[int] = None,
    no_pretty: bool = False,
    plot: Optional[str] = None,
    requirements_locked: Optional[str] = None,
    runtime_environment: Optional[str] = None,
    seed: Optional[int] = None,
    pipeline: Optional[str] = None,
    prescription: Optional[str] = None,
    constraints: Optional[str] = None,
    user_stack_scoring: bool = True,
    dev: bool = False,
    labels: Optional[str] = None,
):
    """Advise package and package versions in the given stack or on solely package only."""
    parameters = locals()
    parameters.pop("click_ctx")

    if pipeline and prescription:
        sys.exit("Options --pipeline/--prescription are disjoint")

    if library_usage:
        if os.path.isfile(library_usage):
            try:
                with open(library_usage, "r") as f:
                    library_usage = json.load(f)
            except Exception:
                _LOGGER.error("Failed to load library usage file %r", library_usage)
                raise
        else:
            library_usage = json.loads(library_usage)

        # Show library usage in the final report.
        parameters["library_usage"] = library_usage

    labels_dict = {}
    if labels:
        if os.path.isfile(labels):
            try:
                with open(labels, "r") as f:
                    labels_dict = json.load(f)
            except Exception:
                _LOGGER.error("Failed to load labels file %r", labels)
                raise
        else:
            labels_dict = json.loads(labels)

        # Show labels in the final report.
        parameters["labels"] = labels_dict

    runtime_environment = RuntimeEnvironment.load(runtime_environment)
    recommendation_type = RecommendationType.by_name(recommendation_type)
    _LOGGER.info("Using recommendation type %s", recommendation_type.name.lower())

    requirements_format = PythonRecommendationOutput.by_name(requirements_format)
    project = _instantiate_project(
        requirements, requirements_locked, runtime_environment=runtime_environment, constraints=constraints
    )

    pipeline_config = None
    if pipeline:
        pipeline_config = PipelineBuilder.load(pipeline)

    parameters["project"] = project.to_dict()
    if pipeline_config is not None:
        parameters["pipeline"] = pipeline_config.to_dict()

    prescription_instance = None
    if prescription:
        if len(prescription) == 1:
            # Click does not support multiple parameters when supplied via env vars. Perform split on delimiter.
            prescription_instance = Prescription.load(*prescription[0].split(","))
        else:
            prescription_instance = Prescription.load(*prescription)

    predictor_class, predictor_kwargs = _get_adviser_predictor(predictor, recommendation_type)
    predictor_kwargs = _get_predictor_kwargs(predictor_config) or predictor_kwargs
    predictor_instance = predictor_class(**predictor_kwargs, keep_history=plot is not None)

    # Use current time to make sure we have possibly reproducible runs - the seed is reported.
    seed = seed if seed is not None else int(time.time())
    _LOGGER.info(
        "Starting resolver using %r predictor with random seed set to %r, predictor parameters: %r",
        predictor_class.__name__,
        seed,
        predictor_kwargs,
    )
    random.seed(seed)
    termial_random.seed(seed)

    resolver = Resolver.get_adviser_instance(
        predictor=predictor_instance,
        project=project,
        labels=labels_dict,
        library_usage=library_usage,
        recommendation_type=recommendation_type,
        limit=limit,
        count=count,
        beam_width=beam_width,
        limit_latest_versions=limit_latest_versions,
        pipeline_config=pipeline_config,
        prescription=prescription_instance,
        cli_parameters=parameters,
    )

    del prescription  # No longer needed, garbage collect it.

    print_func = _PrintFunc(
        partial(
            print_command_result,
            click_ctx=click_ctx,
            analyzer=analyzer_name,
            analyzer_version=analyzer_version,
            output=output,
            pretty=not no_pretty,
        )
    )

    exit_code = subprocess_run(
        resolver,
        print_func,
        plot=plot,
        result_dict={"parameters": parameters},
        with_devel=dev,
        user_stack_scoring=user_stack_scoring,
        verbose=click_ctx.parent.params.get("verbose", False),
    )

    # Push metrics.
    if _THOTH_METRICS_PUSHGATEWAY_URL:
        _METRIC_INFO.labels(_THOTH_DEPLOYMENT_NAME, analyzer_version).inc()
        _METRIC_DATABASE_SCHEMA_SCRIPT.labels(
            analyzer_name, resolver.graph.get_script_alembic_version_head(), _THOTH_DEPLOYMENT_NAME
        ).inc()

        try:
            _LOGGER.debug("Submitting metrics to Prometheus pushgateway %s", _THOTH_METRICS_PUSHGATEWAY_URL)
            push_to_gateway(_THOTH_METRICS_PUSHGATEWAY_URL, job="adviser", registry=prometheus_registry)
        except Exception:
            _LOGGER.exception("An error occurred when pushing metrics")

    click_ctx.exit(int(exit_code != 0))


@cli.command("dependency-monkey")
@click.pass_context
@click.option(
    "--requirements",
    "-r",
    type=str,
    envvar="THOTH_ADVISER_REQUIREMENTS",
    metavar="REQUIREMENTS_FILE",
    required=True,
    help="Requirements to be advised.",
)
@click.option(
    "--requirements-format",
    "-f",
    envvar="THOTH_REQUIREMENTS_FORMAT",
    default="pipenv",
    required=True,
    type=click.Choice(["pipenv", "requirements"]),
    help="The output format of requirements that are computed based on recommendations.",
)
@click.option(
    "--stack-output",
    "-o",
    type=str,
    envvar="THOTH_DEPENDENCY_MONKEY_STACK_OUTPUT",
    metavar="OUTPUT",
    required=True,
    help="Output directory or remote API to print results to, in case of URL a POST request "
    "is issued to the Amun REST API.",
)
@click.option(
    "--library-usage",
    type=str,
    envvar="THOTH_ADVISER_LIBRARY_USAGE",
    help="Add library usage information to dependency-monkey resolution algorithm.",
)
@click.option(
    "--report-output",
    "-R",
    type=str,
    envvar="THOTH_DEPENDENCY_MONKEY_REPORT_OUTPUT",
    metavar="REPORT_OUTPUT",
    required=False,
    default="-",
    help="Output directory or remote API where reports of dependency monkey run should be posted..",
)
@click.option(
    "--seed",
    envvar="THOTH_DEPENDENCY_MONKEY_SEED",
    help="A seed to be used for generating software stack samples (defaults to time if omitted).",
    type=int,
)
@click.option(
    "--count",
    type=int,
    envvar="THOTH_DEPENDENCY_MONKEY_COUNT",
    default=Resolver.DEFAULT_COUNT,
    help="Number of software stacks that should be computed.",
)
@click.option(
    "--decision-type",
    required=False,
    envvar="THOTH_DEPENDENCY_MONKEY_DECISION_TYPE",
    default="all",
    type=click.Choice([e.name.lower() for e in DecisionType]),
    help="A decision type that should be used for generating software stack samples; "
    "if omitted, all software stacks will be created.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    envvar="THOTH_DEPENDENCY_MONKEY_DRY_RUN",
    help="Do not generate software stacks, just report how many software stacks will be "
    "generated given the provided configuration.",
)
@click.option(
    "--context",
    type=str,
    envvar="THOTH_AMUN_CONTEXT",
    metavar="AMUN_JSON",
    help="The context into which computed stacks should be placed (either a file or"
    "a JSON); if omitted, raw software stacks will be created. This option cannot be set when generating "
    "software stacks onto filesystem.",
)
@click.option(
    "--runtime-environment",
    "-e",
    envvar="THOTH_ADVISER_RUNTIME_ENVIRONMENT",
    metavar="ENV_JSON",
    type=str,
    help="Runtime environment specification (file or directly JSON) to describe target environment.",
)
@click.option("--plot", envvar="THOTH_ADVISER_PLOT", type=str, help="Plot predictor history.")
@click.option(
    "--beam-width",
    "-b",
    envvar="THOTH_ADVISER_BEAM_WIDTH",
    type=int,
    default=Resolver.DEFAULT_BEAM_WIDTH,
    help="Width of the beam used.",
)
@click.option(
    "--limit-latest-versions",
    envvar="THOTH_ADVISER_LIMIT_LATEST_VERSIONS",
    type=int,
    default=Resolver.DEFAULT_LIMIT_LATEST_VERSIONS,
    help="Limit number of latest versions considered for dependency graphs.",
)
@click.option("--no-pretty", "-P", is_flag=True, help="Do not print results nicely.")
@click.option(
    "--predictor",
    envvar="THOTH_ADVISER_PREDICTOR",
    default="AUTO",
    type=click.Choice(predictors.__all__ + ["AUTO"]),
    help="Predictor to be used with the resolver.",
)
@click.option(
    "--predictor-config",
    envvar="THOTH_ADVISER_PREDICTOR_CONFIG",
    default=None,
    type=str,
    metavar="CONFIG",
    help="Predictor configuration - passed as a path to YAML file or as a YAML string.",
)
@click.option(
    "--pipeline",
    envvar="THOTH_ADVISER_PIPELINE",
    default=None,
    type=str,
    metavar="PIPELINE",
    help="Pipeline configuration supplied in a form of JSON/YAML or a path to a file, "
    "disjoint with pipeline prescription. Multiple files can be supplied by using `,' as a delimiter.",
)
@click.option(
    "--prescription",
    envvar="THOTH_ADVISER_PRESCRIPTION",
    default=None,
    type=str,
    metavar="PRESCRIPTION",
    help="Pipeline prescription supplied in a form of JSON/YAML or a path to a file, "
    "disjoint with pipeline configuration.",
)
@click.option(
    "--dev/--no-dev",
    envvar="THOTH_ADVISER_DEV",
    is_flag=True,
    default=False,
    show_default=True,
    help="Consider or do not consider development dependencies during resolution.",
)
def dependency_monkey(
    click_ctx: click.Context,
    *,
    beam_width: int,
    count: int,
    decision_type: str,
    predictor: str,
    report_output: str,
    requirements: str,
    requirements_format: str,
    stack_output: str,
    predictor_config: Optional[str] = None,
    context: Optional[str] = None,
    dry_run: bool = False,
    library_usage: Optional[str] = None,
    limit_latest_versions: Optional[int] = None,
    no_pretty: bool = False,
    plot: Optional[str] = None,
    runtime_environment: str = None,
    seed: Optional[int] = None,
    pipeline: Optional[str] = None,
    prescription: Optional[str] = None,
    dev: bool = False,
):
    """Generate software stacks based on all valid resolutions that conform version ranges."""
    parameters = locals()
    parameters.pop("click_ctx")

    if pipeline and prescription:
        sys.exit("Options --pipeline/--prescription are disjoint")

    if library_usage:
        if os.path.isfile(library_usage):
            try:
                with open(library_usage, "r") as f:
                    library_usage = json.load(f)
            except Exception:
                _LOGGER.error("Failed to load library usage file %r", library_usage)
                raise
        else:
            library_usage = json.loads(library_usage)

        # Show library usage in the final report.
        parameters["library_usage"] = library_usage

    runtime_environment = RuntimeEnvironment.load(runtime_environment)
    parameters["runtime_environment"] = runtime_environment.to_dict()

    decision_type = DecisionType.by_name(decision_type)
    requirements_format = PythonRecommendationOutput.by_name(requirements_format)
    project = _instantiate_project(requirements, runtime_environment=runtime_environment)
    parameters["requirements"] = project.pipfile.to_dict()
    parameters["project"] = project.to_dict()

    pipeline_config = None if pipeline is None else PipelineBuilder.load(pipeline)
    if pipeline_config is not None:
        parameters["pipeline"] = pipeline_config.to_dict()

    prescription_instance = None
    if prescription:
        if len(prescription) == 1:
            # Click does not support multiple parameters when supplied via env vars. Perform split on delimiter.
            prescription_instance = Prescription.load(*prescription[0].split(","))
        else:
            prescription_instance = Prescription.load(*prescription)

    # Use current time to make sure we have possibly reproducible runs - the seed is reported.
    seed = seed if seed is not None else int(time.time())
    predictor_class = _get_dependency_monkey_predictor(predictor, decision_type)
    predictor_kwargs = _get_predictor_kwargs(predictor_config)
    predictor_instance = predictor_class(**predictor_kwargs, keep_history=plot is not None)
    _LOGGER.info(
        "Starting resolver using predictor %r with random seed set to %r, predictor parameters: %r",
        predictor_class.__name__,
        seed,
        predictor_kwargs,
    )
    random.seed(seed)
    termial_random.seed(seed)

    resolver = Resolver.get_dependency_monkey_instance(
        predictor=predictor_instance,
        project=project,
        library_usage=library_usage,
        count=count,
        beam_width=beam_width,
        limit_latest_versions=limit_latest_versions,
        decision_type=decision_type,
        pipeline_config=pipeline_config,
        prescription=prescription_instance,
        cli_parameters=parameters,
    )

    del prescription  # No longer needed, garbage collect it.

    context_content = {}
    try:
        with open(context) as f:
            context_content = json.load(f)
    except (FileNotFoundError, IOError):
        # IOError raised if context is too large to be handled with open.
        context_content = json.loads(context)
    parameters["context"] = context_content

    dependency_monkey_runner = DependencyMonkey(
        resolver=resolver,
        stack_output=stack_output,
        context=context_content,
        dry_run=dry_run,
        decision_type=decision_type,
    )

    print_func = _PrintFunc(
        partial(
            print_command_result,
            click_ctx=click_ctx,
            analyzer=analyzer_name,
            analyzer_version=analyzer_version,
            output=report_output,
            pretty=not no_pretty,
        )
    )

    exit_code = subprocess_run(
        dependency_monkey_runner,
        print_func,
        result_dict={"parameters": parameters},
        plot=plot,
        with_devel=dev,
        user_stack_scoring=False,
        # Keep verbose output (stating pipeline units run) in dependency-monkey.
        verbose=True,
    )

    click_ctx.exit(int(exit_code != 0))


@cli.command("validate-prescriptions")
@click.argument("prescriptions", nargs=1, metavar="PRESCRIPTION_DIR")
@click.option(
    "--show-unit-names",
    envvar="THOTH_ADVISER_VALIDATE_PRESCRIPTION_SHOW_UNIT_NAMES",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show names of prescriptions that were validated.",
)
@click.option(
    "--output",
    envvar="THOTH_ADVISER_VALIDATE_PRESCRIPTION_OUTPUT",
    type=str,
    metavar="PRESCRIPTIONS.pickle",
    required=False,
    help="Serialize validated prescriptions into an output pickle file.",
)
def validate_prescription(prescriptions: str, show_unit_names: bool, output: str) -> None:
    """Validate the given prescription."""
    _LOGGER.info("Validating prescriptions in %r", prescriptions)
    prescription = Prescription.validate(prescriptions)
    _LOGGER.info("Prescriptions %r validated successfully", prescriptions)

    result = {
        "prescriptions": [{"name": p[0], "release": p[1]} for p in prescription.prescriptions],
        "count": {
            "boots_count": sum(1 for _ in prescription.iter_boot_units()),
            "pseudonyms_count": sum(1 for _ in prescription.iter_pseudonym_units()),
            "sieves_count": sum(1 for _ in prescription.iter_sieve_units()),
            "steps_count": sum(1 for _ in prescription.iter_step_units()),
            "strides_count": sum(1 for _ in prescription.iter_stride_units()),
            "wraps_count": sum(1 for _ in prescription.iter_wrap_units()),
        },
        "count_all": sum(1 for _ in prescription.units),
    }

    if show_unit_names:
        result["boots"] = sorted(prescription.boots_dict.keys())
        result["pseudonyms"] = sorted(prescription.pseudonyms_dict.keys())
        result["sieves"] = sorted(prescription.sieves_dict.keys())
        result["steps"] = sorted(prescription.steps_dict.keys())
        result["strides"] = sorted(prescription.strides_dict.keys())
        result["wraps"] = sorted(prescription.wraps_dict.keys())

    yaml.safe_dump(result, sys.stdout)

    if output:
        _LOGGER.info("Writing validated prescriptions to %r", output)
        with open(output, "wb") as fp:
            pickle.dump(prescription, fp)


__name__ == "__main__" and cli()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""GitHub release notes pipeline unit."""

import attr

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union
from typing import TYPE_CHECKING
from packaging.specifiers import SpecifierSet

from thoth.adviser.state import State
from voluptuous import Schema
from voluptuous import Required

from .unit import UnitPrescription
from .schema import PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRY_SCHEMA
from .schema import PACKAGE_VERSION_REQUIRED_NAME_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class GHReleaseNotesWrapPrescription(UnitPrescription):
    """GitHub release notes pipeline unit."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): str,
            Required("release_notes"): PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRY_SCHEMA,
            Required("package_version"): PACKAGE_VERSION_REQUIRED_NAME_SCHEMA,
            Required("prescription"): Schema({"run": bool}),
        }
    )

    CONFIGURATION_DEFAULT: Dict[str, Any] = {"package_name": None, "release_notes": None, "package_version": None}

    _configuration = attr.ib(type=Dict[str, Any], kw_only=True, factory=CONFIGURATION_DEFAULT.copy)

    @staticmethod
    def is_wrap_unit_type() -> bool:
        """Check if this unit is of type wrap."""
        return True

    @staticmethod
    def _construct_release_notes_url(
        organization: str, repository: str, tag_version_prefix: Optional[str], locked_version: str
    ) -> str:
        prefix = tag_version_prefix if tag_version_prefix is not None else ""
        return f"https://github.com/{organization}/{repository}/releases/tag/{prefix}{locked_version}"

    @staticmethod
    def _yield_from_resolved_dependencies(
        run_prescription: Dict[str, Any],
        resolved_dependencies_prescription: Union[Dict[str, Any], List[Dict[str, Any]]],
        prescription_conf: Dict[str, Any],
    ) -> Generator[Dict[str, Any], None, None]:
        """Yield configuration based on resolved dependencies prescribed."""
        if isinstance(resolved_dependencies_prescription, list):
            for item in resolved_dependencies_prescription:
                yield {
                    "package_name": item["name"],
                    "release_notes": run_prescription["release_notes"],
                    "package_version": item,
                    "prescription": prescription_conf,
                }
        else:
            yield {
                "package_name": resolved_dependencies_prescription["name"],
                "release_notes": run_prescription["release_notes"],
                "package_version": resolved_dependencies_prescription,
                "prescription": prescription_conf,
            }

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Include this pipeline unit."""
        if not builder_context.is_included(cls):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore

            prescription_conf = {"run": False}
            match = prescription["match"]
            run = prescription["run"]
            if isinstance(match, list):
                for item in match:
                    yield from cls._yield_from_resolved_dependencies(
                        run, item["state"]["resolved_dependencies"], prescription_conf
                    )
            else:
                yield from cls._yield_from_resolved_dependencies(
                    run, match["state"]["resolved_dependencies"], prescription_conf
                )
                return None

        yield from ()
        return None

    def run(self, state: State) -> None:
        """Add release information to justification for selected packages."""
        justification_addition = []
        for resolved_package_tuple in state.resolved_dependencies.values():
            conf_package_version = self.configuration["package_version"]
            if not conf_package_version or resolved_package_tuple[0] != conf_package_version["name"]:
                continue

            version = conf_package_version.get("version")
            if version is not None and resolved_package_tuple[1] not in SpecifierSet(version):
                continue

            develop = conf_package_version.get("develop")
            if develop is not None:
                package_version = self.context.get_package_version(resolved_package_tuple)
                if not package_version:
                    # This is a programming error as the give dependency has to be registered in the context.
                    _LOGGER.error(
                        "No matching package version for %r registered in the context", resolved_package_tuple
                    )
                    continue

                if package_version.develop != develop:
                    continue

            index_url = conf_package_version.get("index_url")
            if not self._index_url_check(index_url, resolved_package_tuple[2]):
                continue

            if self._configuration["prescription"]["run"]:
                # Can happen if prescription states criteria that match multiple times. We add
                # justification only once in such cases.
                continue

            self._configuration["prescription"]["run"] = True

            release_notes_conf = self.configuration["release_notes"]
            justification_addition.append(
                {
                    "type": "INFO",
                    "message": f"Release notes for package {resolved_package_tuple[0]!r}",
                    "link": self._construct_release_notes_url(
                        organization=release_notes_conf["organization"],
                        repository=release_notes_conf["repository"],
                        tag_version_prefix=release_notes_conf.get("tag_version_prefix"),
                        locked_version=resolved_package_tuple[1],
                    ),
                    "package_name": resolved_package_tuple[0],
                }
            )

        state.add_justification(justification_addition)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to notify about environment used."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class EnvironmentInfoBoot(Boot):
    """A boot to notify about environment used."""

    _JUSTIFICATION_LINK_ENV = jl("env")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always."""
        if not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Notify about environment used."""
        runtime_environment = self.context.project.runtime_environment
        base_image = self.context.project.runtime_environment.base_image
        hardware = self.context.project.runtime_environment.hardware
        recommendation_type = (
            self.context.recommendation_type.name.lower() if self.context.recommendation_type else "UNKNOWN"
        )

        self.context.stack_info.extend(
            [
                {
                    "message": f"Using recommendation type {recommendation_type!r}",
                    "link": "https://thoth-station.ninja/recommendation-types/",
                    "type": "INFO",
                },
                {
                    "message": f"Resolving for runtime environment named " f"{runtime_environment.name or 'UNKNOWN'!r}",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "message": f"Resolving for operating system "
                    f"{runtime_environment.operating_system.name!r} "
                    f"in version {runtime_environment.operating_system.version!r}",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "message": f"Resolving for Python version {self.context.project.python_version!r}",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "message": f"Using platform {(runtime_environment.platform or 'UNKNOWN')!r}",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "message": "Using constraints provided"
                    if self.context.project.constraints.package_versions
                    else "No constraints supplied to the resolution process",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "message": "Using supplied static source code analysis"
                    if self.context.library_usage
                    else "No static source code analysis supplied",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "type": "INFO",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using containerized environment {base_image!r}"
                    if base_image
                    else "No containerized environment used",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using CPU family {(hardware.cpu_family or 'UNKNWON')!r} "
                    f"model {(hardware.cpu_model or 'UNKNOWN')!r}",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using CUDA {runtime_environment.cuda_version!r}"
                    if runtime_environment.cuda_version
                    else "No CUDA used",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using cuDNN {runtime_environment.cudnn_version!r}"
                    if runtime_environment.cudnn_version
                    else "No cuDNN used",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using OpenBLAS {runtime_environment.openblas_version!r}"
                    if runtime_environment.openblas_version
                    else "No OpenBLAS used",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using OpenMPI {runtime_environment.openmpi_version!r}"
                    if runtime_environment.openmpi_version
                    else "No OpenMPI used",
                },
                {
                    "type": "INFO",
                    "link": self._JUSTIFICATION_LINK_ENV,
                    "message": f"Using MKL {runtime_environment.mkl_version!r}"
                    if runtime_environment.mkl_version
                    else "No MKL used",
                },
            ]
        )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test providing environment information used in the stack information."""

import pytest

from typing import Any
from typing import Dict
from typing import List

from thoth.common import RuntimeEnvironment
from thoth.python.constraints import Constraints
from thoth.adviser.boots import EnvironmentInfoBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.enums import RecommendationType
from thoth.common import get_justification_link as jl
from ..base import AdviserUnitTestCase


class TestEnvironmentInfoBoot(AdviserUnitTestCase):
    """Test providing environment information used in the stack information."""

    UNIT_TESTED = EnvironmentInfoBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit for adviser runs."""
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

        builder_context.add_unit(self.UNIT_TESTED())

        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    @pytest.mark.parametrize(
        "runtime_environment_dict,expected,use_constraints,use_library_usage",
        [
            (
                {
                    "base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0",
                    "cuda_version": "11.1",
                    "cudnn_version": "7.6.5",
                    "hardware": {"cpu_family": 1, "cpu_model": 2, "gpu_model": "Some GPU model"},
                    "mkl_version": "2021.1",
                    "name": "rhel-9",
                    "openblas_version": "0.3.0",
                    "openmpi_version": "4.0.5",
                    "operating_system": {"name": "rhel", "version": "9"},
                    "platform": "linux-x86_64",
                    "python_version": "3.10",
                    "recommendation_type": "latest",
                },
                [
                    {
                        "link": "https://thoth-station.ninja/recommendation-types/",
                        "message": "Using recommendation type 'latest'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for runtime environment named 'rhel-9'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for operating system 'rhel' in version '9'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for Python version '3.6'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using platform 'linux-x86_64'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using constraints provided",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using supplied static source code analysis",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using containerized environment "
                        "'quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using CPU family 1 model 2",
                        "type": "INFO",
                    },
                    {"link": jl("env"), "message": "Using CUDA '11.1'", "type": "INFO"},
                    {"link": jl("env"), "message": "Using cuDNN '7.6.5'", "type": "INFO"},
                    {"link": jl("env"), "message": "Using OpenBLAS '0.3.0'", "type": "INFO"},
                    {"link": jl("env"), "message": "Using OpenMPI '4.0.5'", "type": "INFO"},
                    {"link": jl("env"), "message": "Using MKL '2021.1'", "type": "INFO"},
                ],
                True,
                True,
            ),
            (
                {},
                [
                    {
                        "link": "https://thoth-station.ninja/recommendation-types/",
                        "message": "Using recommendation type 'latest'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for runtime environment named 'UNKNOWN'",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for operating system None in version None",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Resolving for Python version '3.6'",
                        "type": "INFO",
                    },
                    {"link": jl("env"), "message": "Using platform 'UNKNOWN'", "type": "INFO"},
                    {
                        "link": jl("env"),
                        "message": "No constraints supplied to the resolution process",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "No static source code analysis supplied",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "No containerized environment used",
                        "type": "INFO",
                    },
                    {
                        "link": jl("env"),
                        "message": "Using CPU family 'UNKNWON' model 'UNKNOWN'",
                        "type": "INFO",
                    },
                    {"link": jl("env"), "message": "No CUDA used", "type": "INFO"},
                    {"link": jl("env"), "message": "No cuDNN used", "type": "INFO"},
                    {"link": jl("env"), "message": "No OpenBLAS used", "type": "INFO"},
                    {"link": jl("env"), "message": "No OpenMPI used", "type": "INFO"},
                    {"link": jl("env"), "message": "No MKL used", "type": "INFO"},
                ],
                False,
                False,
            ),
        ],
    )
    def test_run(
        self,
        context: Context,
        runtime_environment_dict: Dict[str, Any],
        expected: List[Dict[str, str]],
        use_constraints: bool,
        use_library_usage: bool,
    ) -> None:
        """Test providing information about the runtime environment."""
        if use_constraints:
            context.project.constraints = Constraints.from_string("flask>=1.3.0")

        if use_library_usage:
            context.library_usage = {"flask": ["flask.App"]}

        context.project.runtime_environment = RuntimeEnvironment.from_dict(runtime_environment_dict)
        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            assert unit.run() is None

        assert context.stack_info == expected
        assert self.verify_justification_schema(context.stack_info)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test sieve for filtering packages coming from a specific index."""

import pytest

from thoth.adviser.sieves import FilterIndexSieve
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserUnitTestCase


class TestFilterIndexSieve(AdviserUnitTestCase):
    """Test package index sieve."""

    UNIT_TESTED = FilterIndexSieve

    @pytest.mark.skip(reason="Index filtering sieve is never registered.")
    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""

    def test_default_configuration(self) -> None:
        """Test obtaining default configuration."""
        unit = FilterIndexSieve()
        assert unit.configuration == {"package_name": None, "index_url": None}

    def test_run_filter(self) -> None:
        """Test filtering a package based on source index used."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        unit = FilterIndexSieve()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "index_url": ["https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple"],
            }
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == []

    def test_run_no_filter(self) -> None:
        """Test not filtering a package based on source index used."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.1.0",
            index=Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple"),
            develop=False,
        )
        unit = FilterIndexSieve()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "index_url": ["https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple"],
            }
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == [package_version]

    def test_run_no_filter_multiple(self) -> None:
        """Test not filtering a package based on source index used."""
        package_version = PackageVersion(
            name="tensorboard",
            version="==2.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        unit = FilterIndexSieve()
        unit.update_configuration(
            {
                "package_name": "tensorboard",
                "index_url": [
                    "https://pypi.org/simple",
                    "https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple",
                ],
            }
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == [package_version]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019-2021 Kevin Postlehtwait, Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Filter out stacks which have require non-existent ABI symbols in Thoth's s2i base image."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class ThothS2IAbiCompatibilitySieve(Sieve):
    """Remove packages if the Thoth's s2i image being used doesn't have necessary ABI."""

    _THOTH_S2I_PREFIX = "quay.io/thoth-station/"
    CONFIGURATION_DEFAULT = {"package_name": None}
    image_symbols = attr.ib(type=Set[str], factory=set, init=False)
    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)

    _LINK = jl("abi_missing")
    _LINK_NO_ABI = jl("no_abi")
    _LINK_BAD_IMAGE = jl("bad_base_image")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register if the base image provided is Thoth's s2i."""
        base_image = builder_context.project.runtime_environment.base_image
        if not builder_context.is_included(cls) and base_image and base_image.startswith(cls._THOTH_S2I_PREFIX):
            yield {}
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize image_symbols."""
        base_image = self.context.project.runtime_environment.base_image
        parsed_base_image = self.get_base_image(base_image, raise_on_error=False)
        if parsed_base_image is None:
            error_msg = (
                f"Cannot determine Thoth s2i version information from {base_image}, "
                "recommendations specific for ABI used will not be taken into account"
            )
            _LOGGER.warning("%s - see %s", error_msg, self._LINK_BAD_IMAGE)
            self.context.stack_info.append(
                {
                    "type": "WARNING",
                    "message": error_msg,
                    "link": self._LINK_BAD_IMAGE,
                }
            )
            self.image_symbols = set()
            super().pre_run()
            return

        thoth_s2i_image_name, thoth_s2i_image_version = parsed_base_image
        self.image_symbols = set(
            self.context.graph.get_thoth_s2i_analyzed_image_symbols_all(
                thoth_s2i_image_name=thoth_s2i_image_name,
                thoth_s2i_image_version=thoth_s2i_image_version,
                is_external=False,
            )
        )
        if not self.image_symbols:
            message = f"No ABI symbols found for {thoth_s2i_image_name!r} in version {thoth_s2i_image_version!r}"
            _LOGGER.warning("%s - see %s", message, self._LINK_NO_ABI)
            self.context.stack_info.append({"type": "WARNING", "message": message, "link": self._LINK_NO_ABI})

        self._messages_logged.clear()
        _LOGGER.debug("Analyzed image has the following symbols: %r", self.image_symbols)
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """If package requires non-present symbols remove it."""
        if not self.image_symbols:
            # No image symbols or the version was not provided, the warning is produced in pre_run method.
            yield from package_versions
            return None

        for pkg_vers in package_versions:
            package_symbols = set(
                self.context.graph.get_python_package_required_symbols(
                    package_name=pkg_vers.name,
                    package_version=pkg_vers.locked_version,
                    index_url=pkg_vers.index.url,
                )
            )

            # Shortcut if package requires no symbols
            if not package_symbols:
                yield pkg_vers
                continue

            missing_symbols = package_symbols - self.image_symbols
            if not missing_symbols:
                yield pkg_vers
            else:
                # Log removed package
                package_tuple = pkg_vers.to_tuple()
                if package_tuple not in self._messages_logged:
                    message = f"Package {package_tuple} was removed due to missing ABI symbols in the environment"
                    _LOGGER.warning("%s - see %s", message, self._LINK)
                    self._messages_logged.add(package_tuple)
                    _LOGGER.debug("The following symbols are not present: %r", missing_symbols)

                continue



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests related to CPU database."""

import pytest

from thoth.adviser.cpu_db import CPUDatabase

from .base import AdviserTestCase


class TestCPUDatabase(AdviserTestCase):
    """Test CPU database."""

    def test_init(self) -> None:
        """Test initializing the database."""
        assert "avx2" in CPUDatabase.get_known_flags()
        assert "avx512" in CPUDatabase.get_known_flags()

    @pytest.mark.parametrize(
        "cpu_family,cpu_model,flag,is_provided",
        [
            (6, int("0x5", base=16), "avx2", True),
            (6, int("0xC", base=16), "avx512", True),
            (999, 0xBAAAAAAD, "avx512", False),
            (12345, 0x8BADF00D, "avx2", False),
            (98765, 0xDEAD10CC, "UNKNOWN_FLAG", False),
        ],
    )
    def test_provides_flag(self, cpu_family: int, cpu_model: int, flag: str, is_provided: bool) -> None:
        """Test if the given CPU provides a flag."""
        assert CPUDatabase.provides_flag(cpu_family, cpu_model, flag) == is_provided



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2022 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap adding a link to the PyTorch index."""

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.wraps import PyTorchReleaseWrap

from ..base import AdviserUnitTestCase


class TestPyTorchReleaseWrap(AdviserUnitTestCase):
    """Test wrap adding a link to the PyTorch index."""

    UNIT_TESTED = PyTorchReleaseWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run_add_justification(self) -> None:
        """Test adding a link to the PyTorch index."""
        state = State()
        package_version_tuple = ("torch", "1.10.2", f"{self.UNIT_TESTED._TORCH_INDEX_PREFIX}cpu")
        state.add_resolved_dependency(package_version_tuple)

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert state.justification == [
            {
                "link": "https://download.pytorch.org/whl/cpu",
                "message": "Package 'torch' in version '1.10.2' is released on the PyTorch index",
                "package_name": "torch",
                "type": "INFO",
            },
        ]
        self.verify_justification_schema(state.justification)

    def test_run_no_justification(self) -> None:
        """Test NOT adding a link to the PyTorch index."""
        state = State()
        state.add_resolved_dependency(("torch", "1.10.2", "https://pypi.org/simple"))

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert not state.justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Sieve packages with CVE when recommendation type is set to security."""

import logging

from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError
from voluptuous import Required
from voluptuous import Schema

from ..enums import RecommendationType
from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class CveSieve(Sieve):
    """Filter out packages with CVEs."""

    CONFIGURATION_DEFAULT = {"package_name": None}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): None,
        }
    )
    _JUSTIFICATION_LINK = jl("cve")

    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)
    _messages_logged_allow_cve = attr.ib(type=Set[Tuple[str, Tuple[str, str, str]]], factory=set, init=False)
    _allow_cves = attr.ib(type=Set[str], factory=set, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Remove CVEs only secure stacks."""
        if (
            builder_context.is_adviser_pipeline()
            and not builder_context.is_included(cls)
            and builder_context.recommendation_type == RecommendationType.SECURITY
        ):
            yield {}
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before running."""
        self._messages_logged.clear()
        self._messages_logged_allow_cve.clear()
        self._construct_allow_cves(self._allow_cves, self.context.labels)
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages with a CVE."""
        for package_version in package_versions:
            try:
                cve_records = self.context.graph.get_python_cve_records_all(
                    package_name=package_version.name,
                    package_version=package_version.locked_version,
                )
            except NotFoundError as exc:
                _LOGGER.warning("Package %r in version %r not found: %r", exc)
                continue

            allow_cves_matched = [cve_record for cve_record in cve_records if cve_record["cve_id"] in self._allow_cves]
            disallowed_cves = [cve_record for cve_record in cve_records if cve_record["cve_id"] not in self._allow_cves]

            package_version_tuple = package_version.to_tuple()

            if not disallowed_cves:
                for allow_cve in allow_cves_matched:
                    messages_logged_allow_cve_key = (allow_cve["cve_id"], package_version_tuple)
                    if messages_logged_allow_cve_key not in self._messages_logged_allow_cve:
                        self._messages_logged_allow_cve.add(messages_logged_allow_cve_key)
                        _LOGGER.warning(
                            "Allowing CVE %r found for package %r based on 'allow-cve' label supplied",
                            allow_cve["cve_id"],
                            package_version_tuple,
                        )

                yield package_version
                continue

            _LOGGER.debug("Found a CVEs for %r: %r", package_version_tuple, cve_records)

            if package_version_tuple not in self._messages_logged:
                self._messages_logged.add(package_version_tuple)
                for cve_record in disallowed_cves:
                    message = (
                        f"Skipping including package {package_version_tuple!r} as a CVE "
                        f"{cve_record['cve_id']!r} was found"
                    )
                    _LOGGER.warning(message)
                    self.context.stack_info.append(
                        {
                            "type": "WARNING",
                            "message": message,
                            "link": cve_record.get("link") or self._JUSTIFICATION_LINK,
                        }
                    )
